/**** N-queens since 2004-09   by Kenji KISE ****/
/**** N-queens OpenMP Version in C           ****/
/************************************************/
#include <u.h>
#include <libc.h>
#include <stdio.h>
/************************************************/
#define NAME "qn24b OpenMP"
#define VER  "version 1.0.0 2004-04-21"
#define MAX 29 /** 32 is a real max! **/
#define MIN 2
#define MAX_TASK 100000

static long long answer[30] = {
    0, 1, 0, 0, 2, 10, 4, 40, 92, 352, 724,
    2680, 14200, 73712, 365596, 2279184,
    14772512, 95815104, 666090624ull,
    4968057848ull, 39029188884ull,
    314666222712ull, 2691008701644ull,
    24233937684440ull,
	227514171973736ull,
    -1
};

void
usage(void)
{
	fprint(2,"usage: queen [size [range]]\n");
	fprint(2,"size is 1 to %d\n",MAX);
	fprint(2,"range is n:m    (Python style)\n");
	fprint(2,"    0 <= n < m <= M\n");
	fprint(2,"you can get M by\n");
	fprint(2,"    queen -n size\n");
	exits("usage");
}

/************************************************/
typedef struct array_t{
  unsigned int cdt; /* candidates        */
  unsigned int col; /* column            */
  unsigned int pos; /* positive diagonal */
  unsigned int neg; /* negative diagonal */
} array;

/** N-queens kernel                            **/
/** n: problem size, h: height, r: row         **/
/************************************************/
long long qn(int n, int h, int r, array *a){
  long long answers = 0;

  for(;;){
    if(r){
      int lsb = (-r) & r;
      a[h+1].cdt = (       r & ~lsb);
      a[h+1].col = (a[h].col & ~lsb);
      a[h+1].pos = (a[h].pos |  lsb) << 1;
      a[h+1].neg = (a[h].neg |  lsb) >> 1;
      
      r = a[h+1].col & ~(a[h+1].pos | a[h+1].neg);
      h++;
    }
    else{
      r = a[h].cdt;
      h--;

      if(h==0) break;
      if(h==n) answers++;
    }
  }
  return answers;
}

/** function to return the time                **/
/************************************************/
long long get_time(void){
  return nsec()/1000;
}

/** print the answer                           **/
/************************************************/
void print_result(int n, long long usec,
                  long long solution){
  int i;
  for(i=0;i<45;i++) printf("="); printf("\n");
  printf("%s %s\n", NAME, VER);
  printf("problem size n        : %d\n", n);
  printf("total   solutions     : %lld\n",
         solution);
  printf("correct solutions     : %lld\n",
         answer[n]);
  printf("million solutions/sec : %.3f\n",
         (double)solution/usec);
  printf("elapsed time (sec)    : %.3f\n",
         usec/1000000.0);
  if(solution!=answer[n])
    printf(" ### Wrong answer\n");
  for(i=0;i<45;i++) printf("="); printf("\n");
  fflush(stdout);
}

/** print the answer                           **/
/************************************************/
void print_partial_result(int n, int j, int k, long long usec,
                  long long solution){
  int i;
  for(i=0;i<45;i++) printf("="); printf("\n");
  printf("%s %s\n", NAME, VER);
  printf("problem size n        : %d\n", n);
  printf("partial solutions     : %lld\n",
         solution);
  printf("range of solutions    : %d:%d\n", j-1,k);
  printf("known total solutions : %lld\n",
         answer[n]);
  printf("million solutions/sec : %.3f\n",
         (double)solution/usec);
  printf("elapsed time (sec)    : %.3f\n",
         usec/1000000.0);
  for(i=0;i<45;i++) printf("="); printf("\n");
  fflush(stdout);
}

/** n-queens solver, specify first 4 level     **/
/************************************************/
long long queen_sub(int n, int *in, int mode){
  int i;
  long long solution=0;   /* solutions          */
  unsigned int row;       /* row candidate      */
  unsigned int h;         /* height or level    */
  array a[MAX]; 

  for(i=0; i<MAX; i++){
    a[i].cdt = a[i].col = a[i].pos = a[i].neg = 0;
  }
  row      = (1<<n)-1;
  a[1].col = (1<<n)-1;
  a[1].pos = 0;
  a[1].neg = 0;
  a[1].cdt = 1; /* care */
  a[2].cdt = 0; /* care */
  
  for(h=1; h<=4; h++){  /** Initialize **/
    int lsb = (1<<in[h]);

    if((a[h].col & lsb)==0) return 0;
    if((a[h].pos & lsb)!=0) return 0;
    if((a[h].neg & lsb)!=0) return 0;

    a[h+1].col = (a[h].col & ~lsb);
    a[h+1].pos = (a[h].pos |  lsb) << 1;
    a[h+1].neg = (a[h].neg |  lsb) >> 1;
    row = a[h+1].col & ~(a[h+1].pos | a[h+1].neg);
  }

  if(mode==0) return 1;

  solution = qn(n, h, row, a);
  return solution;
}

/** get the number of sub problems             **/
/************************************************/
int get_sub_problem_num(int n, int *prob){
  int i;
  int problem=0;
  int max = ((n>>1) + (n&1)) * n * n * n;

  for(i=0; i<=max; i++){
    int in[MAX];

    in[1] = (i-1) / n / n / n;
    in[2] = (i-1) / n / n % n;
    in[3] = (i-1) / n % n;
    in[4] = (i-1) % n;

    if(queen_sub(n, in, 0)){
      problem++;
      prob[problem] = i;
    }
  }
  printf("There are %d tasks\n", problem);
  if(problem > MAX_TASK){
    printf("The number of tasks is too large.\n");
    exits("usage");
  }
  return problem;
}

/************************************************/
long long solver(int n, int id,
                 int problem, int *prob){
  long long ret;
  int p_no = prob[id];
  int in[MAX];

  if(id<1 || id>problem){
    printf("The id %d is out of range.\n", id);
    return 0;
  }

  in[1] = (p_no-1) / n / n / n;
  in[2] = (p_no-1) / n / n % n;
  in[3] = (p_no-1) / n % n;
  in[4] = (p_no-1) % n;

  ret = queen_sub(n, in, 1);
  if(in[1]<(n>>1)) ret = ret * 2;

  return ret;
}

/** main function                              **/
/************************************************/
void
main(int argc, char *argv[]){
  int prob[MAX_TASK]; /* store problem IDs */
  int rets[MAX_TASK]; /* store each answer */
  int i,j,k;
  int n;
  int jobs;
  int nflag=0;
  long long usec    = get_time();
  long long answers = 0;
  char *args[3];

  ARGBEGIN{
  case 'n': nflag++; break;
  default: usage();
  }ARGEND
  if(argc == 0){
	print("number of solutions of n queens problem.\n");
	print("note: equivalent solutions are not removed.\n");
	for(i = 1; answer[i] != -1; i++)
		print("%2d %lld\n",i,answer[i]);
    exits(nil);
  }
  if(argc >2)
	usage();
  if(strchr(argv[0],':'))
	usage();
  n = atoi(argv[0]);
  if(n <= 0 || n > MAX)
    usage();
  jobs = get_sub_problem_num(n, prob);
  if(nflag)
	exits(nil);
  argc--; argv++;
  j = 1; k = jobs;
  if(argc == 1){
	if(getfields(argv[0],args,3,0,":") != 2)
		usage();
	j = atoi(args[0]) + 1;
	k = atoi(args[1]);
  }

#pragma omp parallel for schedule(dynamic)
  for(i=j; i<=k; i++)
    rets[i] = solver(n, i, jobs, prob);

  for(i=j; i<=k; i++) answers += rets[i];
  if(j==1 && k==jobs)
    print_result(n, get_time()-usec, answers);
  else
    print_partial_result(n, j, k, get_time()-usec, answers);
  exits(nil);
}
/************************************************/
