#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <fcall.h>
#include <bio.h>
#include <regexp.h>
#define Extern
#include "exportfs.h"

/*
*	comments for comment
*	$T in comments is a dirname of
*	srvfs -e $objtype/taskfs taskfs $T
*	look also "run"
*/


static int done(char *path);


typedef struct Task		Task;
struct Task{
	char *name; // task name, "task.0" for example
	int ver; // current version, 0,1,2,...
	int done; // done flag
};
static Task *tasks;


/*	Patt defines path pattern and the action
*	the action is invoked only in Xcreate		*/
typedef struct Patt		Patt;
struct Patt{
	char *pat;
	int (*act)(char*);
	Reprog *re;
};

static Patt	patterns[] ={
{"^/task.*/done", done, nil},
{"^/task.*", nil, nil},
{ nil, nil, nil}
};

static int ntasks; // number of tasks
static int nleft; // number of left tasks

char *
strtrim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

/*	we are working on $T where
*	srvfs -e $objtype/taskfs taskfs $T	*/
int
init(void)
{
	int fd;
	Dir *db;
	long i,k,n;
	Patt *pa;
	char buf[80];
	char *args[3];

	/* read dir "tasks" and allocate array "tasks" */
	fd = open("tasks", OREAD);
	if(fd < 0)
		sysfatal("%r");
	n = dirreadall(fd, &db);
	tasks = calloc(n, sizeof(Task));
	for(i = 0; i < n; i++){
		strncpy(buf, db[i].name, sizeof buf);
		if(getfields(buf, args, 3, 0, ".") != 2)
			sysfatal("invalid task name: %s",db[i].name);
		k = atoi(args[1]);
		if(k < 0 || k >= n)
			sysfatal("invalid task no. %s", args[1]);
		tasks[k].name = strdup(db[i].name);
		tasks[k].ver = 0;
		tasks[k].done = 0;
	}
	free(db);
	nleft = ntasks = n;

	/*	reg compile pat to re */
	for(pa = patterns; pa->pat != nil; pa++)
		pa->re = regcomp(pa->pat);

	return 0;
}

/*	path is "/task.2.a/done" for example.
*	we should assign task[2].done=1 and decrease nleft */
static int
done(char *path)
{
	char *p, *args[3];
	char buf[80], buf1[80];
	int k,fd;
	int n;

	n = nleft;
	/* 	don't return here even if n==0
	*	because we must create "dup" marker	in the directory */

	strncpy(buf, path, sizeof buf);
	p = strrchr(buf,'/'); *p = 0;
	p = buf; p++;
	fprint(2,"done %s\n", p);
	snprint(buf1, sizeof buf1,"%s/dup",p);
	getfields(p, args, 3, 0, ".");
	k = atoi(args[1]);
	/*	we might have both "/task.2.a/done" and "/task.2.b/done"
	*	therefore be careful */
	if(tasks[k].done == 0){
		tasks[k].done = 1;
		nleft--;
	}
	else{
		fd = create(buf1,OREAD,0644);
		close(fd);
	}

	if(n > 0 && nleft == 0){
		fprint(2,"all tasks finished\n");
		fd = create("done",OREAD,0644);
		close(fd);
	}
	return 0;
}

/*	path is "/task.2.a/done" for example. */
int
action(char *path)
{
	char *p;
	Patt *pa;

	if(*(path+1) == 0)
		p = "/";
	else
		p = path+1;

	DEBUG(DFD, "checking %s\n", path);

	for(pa = patterns; pa->pat != nil; pa++){
		if(regexec(pa->re, p, nil, 0) == 1){
			DEBUG(DFD, "matched %s\n", path);
			if(pa->act)
				(pa->act)(p);
			return 1;
		}
	}
	return 0;
}

/*	readtaskname() give a taskname to the slave that reads $T/task	*/
int
readtaskname(char *buf, vlong size, vlong offset)
{
	int i,n,v;

	if(offset > 0)
		return 0;
	/* we give a task of minimum version */
	v = 27; n = -1;
	for(i = 0; i < ntasks; i++){
		if(tasks[i].done)
			continue;
		if(tasks[i].ver < v){
			v = tasks[i].ver;
			n = i;
		}
	}
	if(n == -1)
		return 0;
	
	tasks[n].ver++;
	if(tasks[n].ver >= 26)
		return 0;
	/* converting version number 0,1,... to 'a','b',... */
	return snprint(buf,size, "%s.%c",tasks[n].name,'a'+v);
}


int
preaddir(Fid *f, uchar *data, int n, vlong offset)
{
	int r = 0, m;
	Dir *d;

	DEBUG(DFD, "\tpreaddir n=%d wo=%lld fo=%lld\n", n, offset, f->offset);
	if(offset == 0 && f->offset != 0){
		if(seek(f->fid, 0, 0) != 0)
			return -1;
		f->offset = f->cdir = f->ndir = 0;
		free(f->dir);
		f->dir = nil;
	}else if(offset != f->offset){
		werrstr("can't seek dir %lld to %lld", f->offset, offset);
		return -1;
	}

	while(n > 0){
		if(f->dir == nil){
			f->ndir = dirread(f->fid, &f->dir);
			if(f->ndir < 0)
				return f->ndir;
			if(f->ndir == 0)
				return r;
		}
		d = &f->dir[f->cdir++];
		m = convD2M(d, data, n);
		DEBUG(DFD, "\t\tconvD2M %d\n", m);
		if(m <= BIT16SZ){
			DEBUG(DFD, "\t\t\tneeded %d\n", GBIT16(data));
			/* not enough room for full entry; leave for next time */
			f->cdir--;
			return r;
		}else{
			data += m;
			n -= m;
			r += m;
			f->offset += m;
		}
		if(f->cdir >= f->ndir){
			f->cdir = f->ndir = 0;
			free(f->dir);
			f->dir = nil;
		}
	}
	return r;
}
